function tof = formToF ( T, R, X, ssp )

% Compute time of flight (ToF) between target (T) and anchor (A), 
% i.e., arbitrarily two points in ResearchSpace,
%
% ssp: Sound velocity , isogradient sound speed
% slope: sound velocity gradient
% T: transmitterpos
% R: receiverpos
% X: targetpos
%
%

% coeff = polyfit(ssp(:,1),ssp(:,2),1);
% slope = coeff(1);

% 计算导数
dy_dx = diff(ssp(:,1)) ./ diff(ssp(:,2));
differences = diff(dy_dx);

if all(isnan(differences) == 1) && all(isinf(dy_dx)==1)
        %% Constant ssp
%         disp('using distance-based calculate tof')
%         disp('consider constant sound speed profile')
        c = mean( ssp(:,2) );
                t1 = norm( T-X ) ./ c;
                t2 = norm( R-X ) ./ c;
                tof = t1+t2;
elseif all(differences <1e-8) || max(abs(differences))<1500
        %% isogradient ssp 
%         disp('using isogradient-based calculate tof')
%         disp('consider isogradient sound speed profile')
        coeff = polyfit(ssp(:,1),ssp(:,2),1);        
        a =  coeff(1);

                if T(3,:)==X(3,:)              %transmitterpos and targetpos at same depth
                    z_ind = abs(X(3,:));
                    t1 = norm( T-X ) ./ ssp(z_ind);
                else
                    d = norm( T-X );
                    ztx_ind = abs(T(3,:));
                    z_ind    = abs(X(3,:)) ;
                    % ------------ interp1
                    v =  interp1(ssp(:,1),ssp(:,2), ztx_ind);
                    w = interp1(ssp(:,1),ssp(:,2), z_ind);
                    t1 = 1/a .*log( (a*d + sqrt(a^2.*d.^2+4*v.*w) ).^2 / ( 4*v.*w ) );
                end

                if R(3,:)==X(3,:)
                    z_ind = abs(X(3,:));
                    t2 = norm( R-X ) ./ ssp(z_ind);
                else 
                    d = norm( R-X );
                    zrx_ind = abs(R(3,:));
                    z_ind    = abs(X(3,:));
                    % ------------ interp1
                    v =  interp1(ssp(:,1),ssp(:,2), zrx_ind);
                    w = interp1(ssp(:,1),ssp(:,2), z_ind);
                    t2 = 1/a .*log( (a*d + sqrt(a^2.*d.^2+4*v.*w) ).^2 / ( 4*v.*w ) );
                end

                tof = t1+t2;

else
    %% abitrary ssp 
%     disp('using bellhop-based calculate tof')
%     disp('consider abitrary sound speed profile')
            freq = 4e3;
            Nbeams = 0;
            
            sd=[];
            rd=[];
            range=[];
            
            sd = T(3,:);     %source depth   (m)
            rd = X(3,:);     %receiver depth (m)
            range = norm(T-X) / 1e3;       %receiver range (km)
            outFileName = 'ex_arr';     % File name
            Runoption = 'A';                %option solve arr
            bellhopENVgen (ssp, freq, outFileName, sd, rd, range, Nbeams, Runoption);    %write env file
            bellhop ('ex_arr')
            t1 = plotarr('ex_arr.arr',1,1,1);
            
            
            sd = X(3,:);     %source depth   (m)
            rd = R(3,:);     %receiver depth (m)
            range = norm(R-X) / 1e3;       %receiver range (km)
            outFileName = 'ex_arr';     % File name
            Runoption = 'A';                %option solve arr
            bellhopENVgen (ssp, freq, outFileName, sd, rd, range, Nbeams, Runoption);    %write env file
            bellhop ('ex_arr')
            t2 = plotarr('ex_arr.arr',1,1,1);
            
            if isempty(t1) || isempty(t2)
                tof =NaN ;
                disp(['bellhopToF: the candidate point should be cancelled'])
            else
                tof  =  t1 + t2; 
            end
                
end

end
    
%---------- example for bellhopdelay    
% subplot(2,1,1),plotray('ex_arr'), grid on, subplot(2,1,2), plotarr('ex_arr.arr',1,1,1), grid on
% text(0,sd, 'X', 'Color','b' ,'FontSize', 20, 'FontWeight','bold'), hold on, text(range*1e3,rd, 'X', 'Color','k','FontSize', 20,'FontWeight','bold')
