% script_omt_illustration
% yangchaoran @2024
% Mail:yangchaoran_heu@163.com

close all
clc
clear all
addpath('../data','../functions')

%% TODO:  Setup
Ntx = 1;
Nrx = 4;
Ntarget = 2;
%
xint=[1,1500];
yint=[1,1500];
zint=[1,1500];
Edge = [xint;yint;zint];
% ------------ Generate candidate grids coordiate
Ori_interval = 150;%[0.15,0.15,0.15]'*1e3;  %unit  ~m
[x, y, z] = meshgrid( xint(1):Ori_interval:xint(2), yint(1):Ori_interval :yint(2), zint(1):Ori_interval :zint(2) );
Positions = [x(:), y(:), z(:)]'; 

% ------------ Generate fix/random sensors/targets pos
transmitterpos = [1081,481,31]';
receiverpos      = [781,481,331;  931,1231,481;  1081,1381,381;  331,931,531]'; % 针对 Edge=[1.5,1.5,1.5] km case
targetpos0       =  Positions(:,[711,844]);    %[781,481,901;  931,1231,1001]'; %Positions(:,[711,844, 823,]);
% targetpos = targetpos0;
% ------------ When consider offgrid targets pos
bias = 35;    %unit  ~m
targetpos = targetpos0 + bias*ones(3,Ntarget);

% ------------ Plot Setup config
% plot_omtgrid( Positions,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,[] )




%% TODO:  acoustic environment
c = 1540;        % surface sound velocity, unit  ~m/s
N = 30;           % depth sample number
depth = round( linspace( 0,max(zint),N )' ) ;

% ------------ Generate constant sound speed profile
SSP_c= [depth, c *ones(N,1)];
% ------------ Generate isogradient sound speed profile
a = -0.05;
SSP_i = [depth, a*depth + c];
% ------------ Generate abitrary sound speed profile
% load 'exp_ssp'
% SSP_a = SSP;
% ------------ Generate realistic sound speed profile
% load 'SSP_SouthSea'
% SSP_r = SSP;

% ------------ Consider sound velocity with noise 
% vsigm2 = 0;
v_error = 0;
vnoise = [zeros(N,1),v_error*rand(N,1)];%[zeros(N,1),sqrt(vsigm2)*rand(N,1)];
SSP_i0  = SSP_i +vnoise;
% ------------ Plot sound speed profile
% figure;
% plot(SSP_c(:,2),depth, ':^'), hold on, plot(SSP_i0(:,2),depth, ':o'); set(gca,'YDir','reverse'), grid on
% hold on, plot(SSP_a(:,2),SSP_a(:,1),':*'),  hold on,  plot(SSP_r(:,2),SSP_r(:,1),'-s'),
% xlabel('Sound velocity (m/s)'), ylabel('Depth (m)'), legend('Constant SSP','Isogradient SSP','Abitrary SSP','Realistic SSP')
%% TODO:  Generate data
% ------------ Calculate data (ideal/without noise measurement) 
E = [];
for ell = 1:Nrx
    for k = 1:Ntarget
        XYZpos = ( transmitterpos(:) + receiverpos(:,ell) )' / 2;
        tof = formToF( transmitterpos, receiverpos(:,ell), targetpos(:,k), SSP_i ) ;   % SSP_i+vnoise is used for simulation, SSP_r is used for experiment
        E = [E;[transmitterpos', XYZpos, tof]];
    end
end

% ------------ Consider time-delay with noise 
% sigm2 = 10e-5;             % sigma^2
% time_error = 0 + randn(size(E,1),1)*sqrt(sigm2);  measerrordis = time_error*c;
% time_error = ones(size(E,1),1)*0.01;  measerrordis = time_error*c
% E(:,end) = E(:,end) + time_error; 

% ------------ Tips: when there is one measurement is wrong
E(2,7) = 5;


%% TODO:  algorithm1, omt1
%# omt1,standard ellipsoid (distance-based) localization on muti-targets
Q = size( E,1 );               %time-delay measurement, $\Phi_1$ (boundary1)
N = size( Positions,2 );  %candidate grid number, $\Phi_2$     (boundary2)
C = zeros( Q,N );            %cost function matrix
positiondelays = [];
% ------------ Calculate cost function (omt1)
tic

for q=1:Q
    xt = E( 1,1:3 )';
    XYZpos = E( q,4:6 )';
    yr = 2*XYZpos - xt;
    positiondelays = ( vecnorm( Positions-xt ) + vecnorm( Positions-yr ) ) ./ c;
    observeddelay = E(q,7);
    C(q,:) = 0.5*(observeddelay - positiondelays).^2;
end


% ------------ Convex Optimal
ett = ones(N,1);
ettE = ones(Q,1);
cvx_begin
variable xInd(N,1)
variable M(Q,N)
minimize( trace(C'*M) )
subject to
M*ett == ettE;
sum(M)'<=Nrx*xInd;
sum(xInd)<=Ntarget;   
M<=ettE*xInd';
xInd>=0;
M>=0; 
cvx_end
% ------------ Extract 
[amp,ind] = sort(xInd);
pos_inds = ind(end-Ntarget+1:end);
pos_out1 = Positions(:,pos_inds);
diserr1 = plot_omtgrid( Positions,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,pos_out1 )
omt1_time = toc



% ------------ Calculate cost function (omt2)
% -- Only change positiondelays->cost function
tic

for q=1:Q
    xt = E( 1,1:3 )';
    XYZpos = E( q,4:6 )';
    yr = 2*XYZpos - xt;
    for ell=1:size(Positions,2)
        positiondelays(ell) = formToF( xt, yr, Positions(:,ell), SSP_i ) ;%SSP_i
    end
    observeddelay = E(q,7);
    C(q,:) = 0.5*(observeddelay - positiondelays).^2;
end


% ------------ Convex Optimal
ett = ones(N,1);
ettE = ones(Q,1);
cvx_begin
variable xInd(N,1)
variable M(Q,N)
minimize( trace(C'*M) )
subject to
M*ett == ettE;
sum(M)'<=Nrx*xInd;
sum(xInd)<=Ntarget;   
M<=ettE*xInd';
xInd>=0;
M>=0; 
cvx_end
% ------------ Extract 
[amp,ind] = sort(xInd);
pos_inds = ind(end-Ntarget+1:end);
pos_out2 = Positions(:,pos_inds);
diserr2 = plot_omtgrid( Positions,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,pos_out2 )
omt2_time = toc














    
    
%% TODO:  algorithm3, DOTA
G=[];  %update candidate grids set
T = []; %update measurement set
thre_tritau = 0.03;
positiondelays=[];

tic

for q=1:Q
    xt = E( 1,1:3 )';
    XYZpos = E( q,4:6 )';
    yr = 2*XYZpos - xt;
    for ell=1:size(Positions,2)
        positiondelays(ell) = formToF( xt, yr, Positions(:,ell), SSP_i ) ;  %SSP_i
    end

    % ------------ Searching the possible grid
    ind = find( abs( E(q,7)-positiondelays ) < thre_tritau ) ;
    omg = Positions( :,ind);
         % ------------ case: all measure is efficient, update N
    if isempty(omg) == 0
        G = [G,omg]; 
        G = unique(G', 'rows')'; 
        T = [T;E(q,:)];
         % ------------ case: all measure include empty, update N and Q
    else
        disp('measurement error')
    end

end

% ------------ Calculate cost function (DOTA), update Q and N
Q = size( T,1 ); 
N = size( G,2 );
C = zeros( Q,N );  %\tilde{ C }, update cost functioin
positiondelays=[];
PD = [];
for q=1:Q
    xt = T( 1,1:3 )';
    XYZpos = T( q,4:6 )';
    yr = 2*XYZpos - xt;
    for ell=1:size(G,2)
       positiondelays(ell) = formToF( xt, yr, G(:,ell), SSP_i ) ;
    end
    observeddelay = T(q,7);
    C(q,:) = 0.5*(observeddelay - positiondelays).^2;
    
% %     % ------------ modify cost function metric
% %     effG =[];
% %     noneffG=[];
% %     for ell=1:size(G,2)
% %        positiondelays(ell) = formToF( xt, yr, G(:,ell), SSP_i ) ;
% %        v = interp1(SSP_i(:,1),SSP_i(:,2), G(3,ell)); %grid_ell speed
% %        if  v*abs(positiondelays(ell)-T(q,7)) < Ori_interval/4
% %            C(q,ell) = 0.5*(positiondelays(ell)-T(q,7)).^2;
% %            effG = [effG,G(:,ell)];
% %        else
% %            C(q,ell) = 0.5*((positiondelays(ell)-T(q,7)).^2).^0;
% %            noneffG = [noneffG,G(:,ell)];
% %        end
% %     end
     

PD(q,:) = positiondelays;  %SIMO
end
% ------------ Convex Optimal
ett = ones(N,1);
ettE = ones(Q,1);
cvx_begin
variable xInd(N,1)
variable M(Q,N)
minimize( trace(C'*M) )
subject to
M*ett == ettE;
sum(M)'<=Nrx*xInd;
sum(xInd)<=Ntarget;   
M<=ettE*xInd';
xInd>=0;
M>=0; 
cvx_end
% ------------ Extract 
[amp,ind] = sort(xInd);
pos_inds = ind(end-Ntarget+1:end);
pos_out3 = G(:,pos_inds);
diserr3 = plot_omtgrid( G,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,pos_out3 )
Dota_time = toc


%% ------------ Refine estimate version2
sM = M(:,pos_inds); %exploit M matrix of stage1
center = pos_out3;
% zoom_interval = 20; %unit ~m
zoom_interval = thre_tritau*c; %unit ~m
% mu = 5;  %unit ~m
mu = 2*zoom_interval  / 9;  %unit ~m, ϣ10x-,y-,z-
zoomG=[];
for ii = 1:size(center,2)
    xyzint = [center(:,ii)-zoom_interval*ones(3,1) , center(:,ii)+zoom_interval*ones(3,1)];
    [x, y, z] = meshgrid(xyzint(1,1):mu:xyzint(1,2), xyzint(2,1):mu:xyzint(2,2), xyzint(3,1):mu:xyzint(3,2));
    zoomG{ii} = [x(:),y(:),z(:)]';%update grids

   % ------------ Estimate per subregion (search target in target cubic)
    rInd = find(sM(:,ii)>0.5);
    sT = T(rInd',:);   %update measures

    Q = size( sT,1 );
    N = size( zoomG{ii} ,2 ); 
    C = zeros( Q, N );  
    positiondelays=[];
    for q=1:Q
        xt = sT( 1,1:3 )';
        XYZpos = sT( q,4:6 )';
        yr = 2*XYZpos - xt;
       for ell=1:size(zoomG{ii},2)
           positiondelays(ell) = formToF( xt, yr, zoomG{ii}(:,ell), SSP_i ) ;
       end
       observeddelay = sT(q,7);
       C(q,:) = 0.5*(observeddelay - positiondelays).^2;
    end
    % ------------ Convex Optimal
    ett = ones(N,1);
    ettE = ones(Q,1);
    cvx_begin
    variable xInd(N,1)
    variable M(Q,N)
    minimize( trace(C'*M) )
    subject to
    M*ett == ettE;
    sum(M)'<=Nrx*xInd;
    sum(xInd)<=Ntarget;   
    M<=ettE*xInd';
    xInd>=0;
    M>=0; 
    cvx_end
    % ------------ Extract 
    [amp,ind] = sort(xInd);
    pos_inds = ind(end);
    pos_out4(:,ii) = zoomG{ii} (:,pos_inds);
end
zoomG = cell2mat(zoomG);                       zoomG = unique(zoomG', 'rows')';
diserr4 = plot_omtgrid( zoomG,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,pos_out4 )
DotaRefine_time = toc





%% ------------ Plot mismatch time-delay estimate
figure
%
ms = 6;
htrans = plot3(transmitterpos(1,:),transmitterpos(2,:),transmitterpos(3,:),'sk', 'MarkerSize',ms, 'LineWidth', 1.5);
hold on
hres = plot3(receiverpos(1,:),receiverpos(2,:),receiverpos(3,:),'vk', 'MarkerSize',ms, 'LineWidth', 1.5);  
hold on
htarget = plot3(targetpos(1,:),targetpos(2,:),targetpos(3,:),'om', 'MarkerSize',ms, 'LineWidth', 1.5);
hold on
hgrid = plot3(Positions(1,:),Positions(2,:),Positions(3,:),'g.');
hold on
set(gca, 'ZDir', 'reverse'), grid on

%
hest1 = plot3(pos_out1(1,:),pos_out1(2,:),pos_out1(3,:), 'b+', 'MarkerSize',ms, 'LineWidth', 1.5);  %OMT_1 estimate
hest2 = plot3(pos_out2(1,:),pos_out2(2,:),pos_out2(3,:), 'r+', 'MarkerSize',ms, 'LineWidth', 1.5);  %OMT_2 estimate
hest4 = plot3(pos_out4(1,:),pos_out4(2,:),pos_out4(3,:), 'k+', 'MarkerSize',ms, 'LineWidth', 1.5);   %DOTA estimate
hvec1 = [htrans,hres,htarget,hgrid,hest1,hest2,hest4]';            
strcell1 = {'transmitter','receivers','targets','grid','OMT_1','OMT_2','DOTA'};
legend(hvec1,strcell1);
xlabel('x-pos'),ylabel('y-pos'),zlabel('z-pos')







%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%










% %% ------------ Refine estimate
% center = pos_out;
% zoom_interval = 20; %unit ~m
% mu = 5;  %unit ~m
% zoomG=[];
% pos_out3=[];
% for ii = 1:size(center,2)
%     xyzint = [center(:,ii)-zoom_interval*ones(3,1) , center(:,ii)+zoom_interval*ones(3,1)];
%     [x, y, z] = meshgrid(xyzint(1,1):mu:xyzint(1,2), xyzint(2,1):mu:xyzint(2,2), xyzint(3,1):mu:xyzint(3,2));
%     zoomG{ii} = [x(:),y(:),z(:)]';
% % ------------ Estimate per subregion
%     Q = size( T,1 ); 
%     N = size( zoomG{ii} ,2 );
%     C = zeros( Q,N );  %\tilde{ C }, update cost functioin
%     positiondelays=[];
%     
% 
%     for q= 2:2:Q%1:Q
%         xt = T( 1,1:3 )';
%         XYZpos = T( q,4:6 )';
%         yr = 2*XYZpos - xt;
%         for ell=1:size(zoomG{ii},2)
%            positiondelays(ell) = formToF( xt, yr, zoomG{ii}(:,ell), SSP_i ) ;
%         end
%         observeddelay = T(q,7);
%         C(q,:) = 0.5*(observeddelay - positiondelays).^2;
%     end
% 
%     % ------------ Constrant cost function decrease coupling influence
% %          C(C>0.01)=0;  %0.02 is constrant distance between ajact target, depend on bias
%          title('DOTA')
%     % ------------ Convex Optimal
%     ett = ones(N,1);
%     ettE = ones(Q,1);
%     cvx_begin
%     variable xInd(N,1)
%     variable M(Q,N)
%     minimize( trace(C'*M) )
%     subject to
%     M*ett == ettE;
%     sum(M)'<=Nrx*xInd;
%     sum(xInd)<=Ntarget;   
%     M<=ettE*xInd';
%     xInd>=0;
%     M>=0; 
%     cvx_end
% 
%     % ------------ Extract 
%     [amp,ind] = sort(xInd);
%     pos_inds = ind(end);
%     pos_out3(:,ii) = zoomG{ii} (:,pos_inds);
% 
%     figure
%     plot3(  (1:size(M,2) )'  * ones(1,Q),    ones(size(M,2),1) * (1:Q),    M'  )
%     xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel('transport plan weight (M)'), grid on
% 
% 
% end
% zoomG = cell2mat(zoomG);                       zoomG = unique(zoomG', 'rows')';
% diserr4 = plot_omtgrid( zoomG,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,pos_out3 )
% 
% 

% hold on, 
%     hest3 = plot3(pos_out(1,:),pos_out(2,:),pos_out(3,:), 'r+', 'MarkerSize',ms, 'LineWidth', 1.5);
%     hvec1 = [htrans,hres,htarget,hgrid,hest1,hest2,hest3]';            
%     strcell1 = {'transmitter','receivers','targets','grid','omt_1','omt_2','DOTA'};
%     legend(hvec1,strcell1);
%     xlabel('x-pos'),ylabel('y-pos'),zlabel('z-pos')













%% Test
%%
% ------------ test1: why zoom is wrong
% figure;
% Xc = zeros(Q,size(C,2));
% Yc = zeros(Q,size(C,2));
% Zc = zeros(Q,size(C,2));
% 
% for q=1:Q
%         x = (1:size(C,2) )';              Xc(q,:) = x;
%         y = q*ones(size(C,2),1);    Yc(q,:) = y;
%         z = 1./C(q,:);
%         zz = C(q,:);                        Zc(q,:) =zz;
% %         plot3(x,y,z), hold on, grid on
% end
% xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel('inv of cost function (1./C)')
% 
% figure
% Zc(Zc>0.03)=100;
% plot3(Xc',Yc',Zc'),grid on
% xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel(' cost function (C)')
% 
% figure
% plot3(Xc',Yc',1./Zc'),grid on
% xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel('inv of cost function (1./C)')


%%
% ------------ test2: constrant C how influence M
%     ett = ones(N,1);
%     ettE = ones(Q,1);
%     cvx_begin
%     variable kInd(N,1)
%     variable Zm(Q,N)
%     minimize( trace(Zc'*Zm) )
%     subject to
%     Zm*ett == ettE;
%     sum(Zm)'<=Nrx*kInd;
%     sum(kInd)<=Ntarget;   
%     Zm<=ettE*kInd';
%     kInd>=0;
%     Zm>=0; 
%     cvx_end
% [amp,ind] = sort(kInd);
% pos_inds = ind(end-Ntarget+1:end);
% pos_out_Tips= G(:,pos_inds);
% diserr_Tips = plot_omtgrid( G,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,pos_out_Tips )
% figure
% plot3(  (1:size(Zm,2) )'  * ones(1,Q),    ones(size(Zm,2),1) * (1:Q),    Zm'  )
% xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel('transport plan weight (M)'), grid on


%%
% ------------ test3: modify C how influence M  (non finished, wrong code)
% % %     effG =[];
% % %     noneffG=[];
% % % 
% % %     ssp = SSP_a;
% % %  for q=1:Q
% % %     for ell=1:size(G,2)
% % %        pd(ell) = formToF( xt, yr, G(:,ell), ssp ) ;
% % %        v = interp1(ssp(:,1),ssp(:,2), G(3,ell));                  %grid_ell speed
% % %        if  v*abs(pd(ell)-T(q,7)) < Ori_interval/2
% % %            Zc(q,ell) = 0.5*(pd(ell)-T(q,7)).^2;
% % %            effG = [effG,G(:,ell)];
% % %        else
% % %            Zc(q,ell) = 0.5*((pd(ell)-T(q,7)).^2).^0;
% % %            noneffG = [noneffG,G(:,ell)];
% % %        end
% % %     end
% % %  end
% % %     plot_omtgrid( effG,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,[] )
% % %     plot_omtgrid( noneffG,transmitterpos,receiverpos,targetpos,Nrx,Ntarget,[] )
% % % figure;
% % % plot3(  (1:size(Zc,2) )'  * ones(1,Q),    ones(size(Zc,2),1) * (1:Q),    1./Zc'  )
% % % xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel('inv of cost function (1./C)')





%%
% ------------ test2: why zoom is wrong
% figure;
% Xm = zeros(Q,size(C,2));
% Ym = zeros(Q,size(C,2));
% Zm = zeros(Q,size(C,2));
% for q=1:Q
%         x = (1:size(M,2) )';               Xm(q,:) = x;
%         y = q*ones(size(M,2),1);      Ym(q,:) = y;
%         z = M(q,:);                             Zm(q,:) =z;
%         plot3(x,y,z), hold on, grid on
% end
% xlabel('gridNumber (N)'), ylabel('measurenetNumber (Q)'), zlabel('transport plan weight (M)')
% 
% figure
% plot3(Xm',Ym',Zm'),grid on
% 
% 
% figure;
% plot(xInd), xlabel('gridNumber'), ylabel('distribution (f)')